import { Observable, Subject } from "rxjs";
import dotenv from 'dotenv';
import { v4 as uuidv4 } from 'uuid';
import { WebsocketTransportService } from "./websocket";
import { ITransportReceiving, ITransportSuper, ITransportTransmitting, ReceiverProfile, TransportEventNotification, TransportMessage, TransportSettings, TransportType } from "../interface/ITransport.interface";
import { error } from "console";
dotenv.config();
/* This transport manager will be instantiating the necessary transport to deal with tranmission and receiving from different receivers
So how?: */
export class TransportManager {
    private incomingMessage: Subject<TransportMessage> = new Subject()
    private outGoingMessage: Subject<TransportMessage> = new Subject()
    private responseMessage: Subject<FisMessage> = new Subject()
    private connectedClients: ReceiverProfile[] = []
    // this is assuming that the application will do request response, otherwise, just one will do.
    private transportService!: ITransportSuper
    private transmittingService!: ITransportTransmitting
    private receivingService!: ITransportReceiving
    private serverPort!: number
    private serverUrl!: string

    constructor(port?: number, url?: string) {
        /* For web browser ui, because this server can also act as a receiver, not just transmitter as well.  */
        if (port) {
            this.serverPort = port
            this.establishTransportTransmitting(process.env.Transport as TransportType).then((transmission: ITransportTransmitting) => {
                transmission.getEventNotification().subscribe((notification: TransportEventNotification) => {
                    if (notification.event == `Connection` && notification.description == 'New Client Connected') {
                        this.connectedClients.push(notification.data)
                        this.handleNewReceiver(notification.data)
                    }
                })
            }).catch((error) => console.error(error))
        }
        if (url) {
            this.serverUrl = url
            this.establishTransportReceiving(process.env.Transport as TransportType).then((receivingService: ITransportReceiving) => {
                // console.log(receivingService.getInfo('receiver'))
                receivingService.receive().subscribe(this.incomingMessage)
            })
        }
        if (!port && !url) console.error(`No role has assigned for transport service due to lack of information provided!`)
        this.processOutputMessage()
    }

    public send(message: FisMessage): void {
        this.responseMessage.next(message)
    }

    private async establishTransportTransmitting(tranportType: TransportType): Promise<ITransportTransmitting> {
        // console.log(`Setting up ${tranportType.transportName} transmitter`)
        return new Promise((resolve, reject) => {
            let setting: TransportSettings = {
                profileInfo: {
                    id: uuidv4(),
                    name: 'Server',
                    port: this.serverPort,
                    transport: tranportType
                }
            }
            if (tranportType == 'WEBSOCKET') {
                this.transportService = new WebsocketTransportService(setting)
                this.transmittingService = this.transportService
                resolve(this.transmittingService)
            } else {
                reject(`No such Transport Type Exist...`)
            }
        })
    }

    private async establishTransportReceiving(tranportType: TransportType): Promise<any> {
        return new Promise((resolve, reject) => {
            // will be sharing one instance, This part may break the code
            if (tranportType == 'WEBSOCKET') {
                this.receivingService = this.transportService
                resolve(this.receivingService)
            }
        })
    }

    // This function will wrap the message to be delivered over to established transport to be sent over
    private processOutputMessage(): void {
        this.responseMessage.subscribe({
            next: (message: FisMessage) => {
                if (message.header.messageName == 'NotificationMessage') {
                    // This is just here temporarily. Because The application will be concerned with whom the party is subcribed to, not this transport manager
                    this.outGoingMessage.next({
                        id: '',
                        receiverID: '',
                        payload: message,
                        event: `notification`
                    })
                }
                if (message.header.messageName == 'ResponseMessage') {
                    // Map responses according to the requestID???
                    this.outGoingMessage.next({
                        id: '',
                        receiverID: '',
                        payload: message,
                        event: `resposne`
                    })
                }
            }
        })
    }

    private transmitWrappedMessages(): void {
        this.outGoingMessage.subscribe(message => {
            this.transmittingService.transmit(message)
        })
    }

    // Responsibility include subscribing to event notification as well as tranpport instance for message transmission
    private handleNewReceiver(receiver: ReceiverProfile) {
        receiver.eventNotification.subscribe({
            next: event => {
                // new request will be handled. And then manager will here will pick it up from eventNotification and respond accordingly if there's a need for it
            },
            error: error => console.error(error)
        })
    }
}


export interface FisMessage {
    header: {
        messageName: 'RequestMessage' | 'ResponseMessage' | 'NotificationMessage',
        messageID: string
    },
    data: any
}