import { Observable, Subject } from "rxjs";
import { Socket, Socket as ClientSocket } from "socket.io-client";
import { Socket as SocketForConnectedClient } from "socket.io"
import { startClientSocketConnection, startSocketServer } from "../utils/socket.utils";
import { ITransportReceiving, ITransportTransmitting, ReceiverProfile, TransmitterProfile, TransportEventNotification, TransportMessage, TransportSettings } from "../interface/ITransport.interface";
import { v4 as uuidv4 } from 'uuid'

/* Just code in the context that this websocket service will be handling multiple UI clients. Can think about the server communication at a later time. */
export class WebsocketTransportService implements ITransportTransmitting, ITransportReceiving {
    private transmitterProfile!: TransmitterProfile | null
    private socketReceiverProfile: ReceiverProfile[] = []
    private clientSocket!: ClientSocket
    private incomingMessage: Subject<TransportMessage> = new Subject()
    private eventNotification: Subject<TransportEventNotification> = new Subject()

    constructor(setting: TransportSettings) {
        startSocketServer(setting.profileInfo.port as number).subscribe({
            next: (connectedClient: SocketForConnectedClient) => {
                this.handleNewSocketClient(connectedClient)
            },
            error: error => console.error(error),
        })
    }

    // for transmission(Server Only, not applicable for client Socket)
    public async transmit(message: TransportMessage): Promise<string> {
        return new Promise((resolve, reject) => {
            let receiverInstance = this.socketReceiverProfile.find(obj => obj.uuid === message.receiverID);
            if (receiverInstance) {
                (receiverInstance?.instance as SocketForConnectedClient).emit(message.event, message.payload)
                resolve('Message called for delivery...')
            } else {
                reject(`Receiver cannot be found or doesn't exist...`)
            }
        })
    }

    public getEventNotification(): Observable<TransportEventNotification> {
        return this.eventNotification as Observable<TransportEventNotification>
    }

    // Set up socket listeners to start listening for different events
    private handleNewSocketClient(socket: SocketForConnectedClient): void {
        console.log(`Setting up listeners for socket:${socket.id}`)
        // returns the socket client instance 
        let receiverProfile: ReceiverProfile = {
            uuid: uuidv4(),
            name: `Client`,
            dateCreated: new Date(),
            transportType: `WEBSOCKET`,
            eventNotification: new Subject(),
            instance: socket
        }
        this.eventNotification.next({ event: 'Connection', description: 'New Client Connected', transportType: `WEBSOCKET`, data: receiverProfile })
        this.socketReceiverProfile.push(receiverProfile)

        /* Generally, we don't need this unless in the case of being the receiver */
        socket.on('message', (message: any) => {
            // here
        })

        socket.on('request', (request: any) => {
            // here
        })

        socket.on('notification', (notification: any) => {
            // logic here
        })

        socket.on('disconnect', () => {
            receiverProfile.eventNotification.next({ event: 'Connection', description: `Existing Client ${socket.id} disonnected`, transportType: `WEBSOCKET`, data: receiverProfile })
        })
    }


    // for client UI eg
    public receive(): Observable<any> {
        return this.incomingMessage.asObservable()
    }
}
